% GA Search

clear all
close all
clc
format long

load er_500_10_nets
load er_500_10_opt
load er_500_10_all_sols

nLabels=1:N;

mu=0.5;        % mutation rate
N=10;           % population size
s=1;            % selection coefficient
ts=1000;        % time steps
O=zeros(20*20,17);
mm=NNN/2;

rr=1;
for n=1:20
    
    E=rE{n};    % distances
    A=rA{n};    % adjacency matrix
    W=rW{n};    % weighted adjacency matrix
    D=rD{n};    % shortest network distances between each point pair
    DS=rDS{n};  % shortest network distance from each point to focal node
    ND=rND{n};  % set of distant nodes
    NC=rNC{n};  % set of close nodes
    DC=rDC{n};  % degree of each node
    CC=rCC{n};  % closeness centrality of each node
    BC=rBC{n};  % betweenness centrality of each node
    EC=rEC{n};  % eigenvector centrality of each node
    PR=rPR{n};  % pagerank of each node
    F=rF{n};    % focal node index
    T=rT{n};    % threshold
    MaxC=max(AE{n}(:,3));    % max cost for scaling
    MaxB=max(AE{n}(:,4));    % max benefit
    
    % index and order nodes by characteristics
    IO=[nLabels' DC CC BC EC PR DS];
    ICDC=sortrows(IO(NC,[1 2]),2,'ascend');
    IDDC=sortrows(IO(ND,[1 2]),2,'ascend');
    UCDC=unique(ICDC(:,2));
    UDDC=unique(IDDC(:,2));
    ICCC=sortrows(IO(NC,[1 3]),2,'ascend');
    IDCC=sortrows(IO(ND,[1 3]),2,'ascend');
    ICBC=sortrows(IO(NC,[1 4]),2,'ascend');
    IDBC=sortrows(IO(ND,[1 4]),2,'ascend');
    ICEC=sortrows(IO(NC,[1 5]),2,'ascend');
    IDEC=sortrows(IO(ND,[1 5]),2,'ascend');
    ICPR=sortrows(IO(NC,[1 6]),2,'ascend');
    IDPR=sortrows(IO(ND,[1 6]),2,'ascend');
    ICDS=sortrows(IO(NC,[1 7]),2,'ascend');
    IDDS=sortrows(IO(ND,[1 7]),2,'ascend');
    
    for r=1:20
        % random numbers for mutations
        MC1=rand(N,ts);
        MC2=randi(5,N,ts);
        MC=(MC1<mu).*MC2;
        MD1=rand(N,ts);
        MD2=randi(5,N,ts);
        MD=(MD1<mu).*MD2;
        
        tic
        % GENETIC ALGORITHM MAIN PROGRAM
        
        % initialize close and distant populations
        P=zeros(N,4);
        %     P(:,1)=randsample(NC,N);
        for i=1:N
            P(i,1)=randsample(ND,1);
            test=NC(find(E(P(i,1),NC)+DS(NC)'<=T));
            while isempty(test)==1
                P(i,1)=randsample(ND,1);
                test=NC(find(E(P(i,1),NC)+DS(NC)'<=T));
            end
            P(i,2)=test(randsample(length(test),1));
        end
        
        k=1;
        for i=1:ts
            
            % calculate costs and benefits
            for j=1:N
                P(j,3)=E(P(j,1),P(j,2));
                Ap=A;
                Ap(P(j,1),P(j,2))=1;
                Ap(P(j,2),P(j,1))=1;
                DSp=distances(graph(Ap.*E),nLabels,F);
                P(j,4)=length(find(DSp<=T))-length(find(DS<=T));
            end
            
            % selection
            if sum(P(:,4))~=0
                fit=P(:,4)/sum(P(:,4));
                S=randsample(N,N,true,fit);
                P=P(S,:);
            end
            
            % termination criteria
            if max(P(:,4))==MaxB
                break
            end
            
            for j=1:N
                
                if MD(j,k)>0
                    
                    if MD(j,k)==1       % shuffle degree
                        pos=find(IDDC(:,1)==P(j,1));
                        pos2=find(UDDC==IDDC(pos,2));
                        P(j,1)=IDDC(randsample(find(IDDC(:,2)==UDDC(pos2)),1));
                    elseif MD(j,k)==2       % decrease degree
                        pos=find(IDDC(:,1)==P(j,1));
                        if IDDC(pos,2)>min(UDDC)
                            pos2=find(UDDC==IDDC(pos,2));
                            P(j,1)=IDDC(randsample(find(IDDC(:,2)==UDDC(pos2-1)),1));
                        end
                    elseif MD(j,k)==3   % increase degree
                        pos=find(IDDC(:,1)==P(j,1));
                        if IDDC(pos,2)<max(UDDC)
                            pos2=find(UDDC==IDDC(pos,2));
                            P(j,1)=IDDC(randsample(find(IDDC(:,2)==UDDC(pos2+1)),1));
                        end
                    elseif MD(j,k)==4   % decrease distance
                        pos=find(IDDS(:,1)==P(j,1));
                        if pos>1
                            P(j,1)=IDDS(pos-1);
                        end
                    elseif MD(j,k)==5   % increase distance
                        pos=find(IDDS(:,1)==P(j,1));
                        if pos<250
                            P(j,1)=IDDS(pos+1);
                        end
%                     elseif MD(j,k)==6   % decrease closeness
%                         pos=find(IDCC(:,1)==P(j,1));
%                         if pos>1
%                             P(j,1)=IDCC(pos-1);
%                         end
%                     elseif MD(j,k)==7   % increase closeness
%                         pos=find(IDCC(:,1)==P(j,1));
%                         if pos<250
%                             P(j,1)=IDCC(pos+1);
%                         end
%                     elseif MD(j,k)==8   % decrease betweenness
%                         pos=find(IDBC(:,1)==P(j,1));
%                         if pos>1
%                             P(j,1)=IDBC(pos-1);
%                         end
%                     elseif MD(j,k)==9   % increase betweenness
%                         pos=find(IDBC(:,1)==P(j,1));
%                         if pos<250
%                             P(j,1)=IDBC(pos+1);
%                         end
%                     elseif MD(j,k)==10   % decrease eigenvector
%                         pos=find(IDEC(:,1)==P(j,1));
%                         if pos>1
%                             P(j,1)=IDEC(pos-1);
%                         end
%                     elseif MD(j,k)==11   % increase eigenvector
%                         pos=find(IDEC(:,1)==P(j,1));
%                         if pos<250
%                             P(j,1)=IDEC(pos+1);
%                         end
%                     elseif MD(j,k)==12   % decrease pagerank
%                         pos=find(IDPR(:,1)==P(j,1));
%                         if pos>1
%                             P(j,1)=IDPR(pos-1);
%                         end
%                     elseif MD(j,k)==13   % increase pagerank
%                         pos=find(IDPR(:,1)==P(j,1));
%                         if pos<250
%                             P(j,1)=IDPR(pos+1);
%                         end
                    end
                    
                end
                
                if MC(j,k)>0
                    
                    if MC(j,k)==1       % shuffle degree
                        pos=find(ICDC(:,1)==P(j,2));
                        pos2=find(UCDC==ICDC(pos,2));
                        P(j,2)=ICDC(randsample(find(ICDC(:,2)==UCDC(pos2)),1));
                    elseif MC(j,k)==2       % decrease degree
                        pos=find(ICDC(:,1)==P(j,2));
                        if ICDC(pos,2)>min(UCDC)
                            pos2=find(UCDC==ICDC(pos,2));
                            P(j,2)=ICDC(randsample(find(ICDC(:,2)==UCDC(pos2-1)),1));
                        end
                    elseif MC(j,k)==3   % increase degree
                        pos=find(ICDC(:,1)==P(j,2));
                        if ICDC(pos,2)<max(UCDC)
                            pos2=find(UCDC==ICDC(pos,2));
                            P(j,2)=ICDC(randsample(find(ICDC(:,2)==UCDC(pos2+1)),1));
                        end
                    elseif MC(j,k)==4   % decrease distance
                        pos=find(ICDS(:,1)==P(j,2));
                        if pos>1
                            P(j,2)=ICDS(pos-1);
                        end
                    elseif MC(j,k)==5   % increase distance
                        pos=find(ICDS(:,1)==P(j,2));
                        if pos<250
                            P(j,2)=ICDS(pos+1);
                        end
%                     elseif MC(j,k)==6   % decrease closeness
%                         pos=find(ICCC(:,1)==P(j,2));
%                         if pos>1
%                             P(j,2)=ICCC(pos-1);
%                         end
%                     elseif MC(j,k)==7   % increase closeness
%                         pos=find(ICCC(:,1)==P(j,2));
%                         if pos<250
%                             P(j,2)=ICCC(pos+1);
%                         end
%                     elseif MC(j,k)==8   % decrease betweenness
%                         pos=find(ICBC(:,1)==P(j,2));
%                         if pos>1
%                             P(j,2)=ICBC(pos-1);
%                         end
%                     elseif MC(j,k)==9   % increase betweenness
%                         pos=find(ICBC(:,1)==P(j,2));
%                         if pos<250
%                             P(j,2)=ICBC(pos+1);
%                         end
%                     elseif MC(j,k)==10   % decrease eigenvector
%                         pos=find(ICEC(:,1)==P(j,2));
%                         if pos>1
%                             P(j,2)=ICEC(pos-1);
%                         end
%                     elseif MC(j,k)==11   % increase eigenvector
%                         pos=find(ICEC(:,1)==P(j,2));
%                         if pos<250
%                             P(j,2)=ICEC(pos+1);
%                         end
%                     elseif MC(j,k)==12   % decrease pagerank
%                         pos=find(ICPR(:,1)==P(j,2));
%                         if pos>1
%                             P(j,2)=ICPR(pos-1);
%                         end
%                     elseif MC(j,k)==13   % increase pagerank
%                         pos=find(ICPR(:,1)==P(j,2));
%                         if pos<250
%                             P(j,2)=ICPR(pos+1);
%                         end
                    end
                    
                end
            end
            
            k=k+1;
        end
        t=toc;
        
        CB=P;
        o=CB(find(CB(:,4)==max(CB(:,4))),:);
        if length(o)>1
            o=o(find(o(:,3)==min(o(:,3))),:);
            o=o(1,:);
        end
        
        %     AE{n}=[CB DC(CB(:,1)) DC(CB(:,2)) CC(CB(:,1)) CC(CB(:,2)) BC(CB(:,1)) BC(CB(:,2)) EC(CB(:,1)) EC(CB(:,2)) PR(CB(:,1)) PR(CB(:,2)) DS(CB(:,1)) DS(CB(:,2))];
        
        O(rr,:)=[t o(1) o(2) o(3) o(4) DC(o(1)) DC(o(2)) CC(o(1)) CC(o(2)) BC(o(1)) BC(o(2)) EC(o(1)) EC(o(2)) PR(o(1)) PR(o(2)) DS(o(1)) DS(o(2)) KC(o(1)) KC(o(2)) NM(o(1)) NM(o(2)) KG(o(1)) KG(o(2))];
        if (rr>1) && (mod(rr-1,20)~=0)
            O(rr,1)=t-O(rr-1,1);
        end
        
        [rr O(rr,1) o(3) o(4)]
        
        rr=rr+1;
        
    end
    
end

MO=zeros(ln,6);
for i=1:ln
    MO(i,:)=[mean(O((rrr*(i-1)+1):(rrr*i),1)) var(O((rrr*(i-1)+1):(rrr*i),1)) mean(O((rrr*(i-1)+1):(rrr*i),4)) var(O((rrr*(i-1)+1):(rrr*i),4)) mean(O((rrr*(i-1)+1):(rrr*i),5)) var(O((rrr*(i-1)+1):(rrr*i),5))];
end

save(['er_',num2str(N),'_GA.mat'],'O','MO')






