% Adaptive Neighborhood Hill Climbing Search

% clear all
% for NNN=[500 1000 2000]
%     HCVN(NNN)
% end

function HCVN(NNN)
% clear all
% NNN=500;
close all
clc
format long

load(['er_',num2str(NNN),'_nets.mat']);

nLabels=1:N;
ln=length(rA);
rr=1;
O=zeros(ln,23);
rrr=20;     % number of restarts
mm=NNN/2;

for n=1:ln
    
    E=rE{n};    % distances
    A=rA{n};    % adjacency matrix
    W=rW{n};    % weighted adjacency matrix
    D=rD{n};    % shortest network distances between each point pair
    DS=rDS{n};  % shortest network distance from each point to focal node
    ND=rND{n};  % set of distant nodes
    NC=rNC{n};  % set of close nodes
    DC=rDC{n};  % degree of each node
    CC=rCC{n};  % closeness centrality of each node
    BC=rBC{n};  % betweenness centrality of each node
    EC=rEC{n};  % eigenvector centrality of each node
    PR=rPR{n};  % pagerank of each node
    KC=rKC{n};
    NM=rNM{n};
    KG=rKG{n};
    
    F=rF{n};    % focal node index
    T=rT{n};    % threshold
    %     MaxC=max(AE{n}(:,3))    % max cost for scaling
    %     MaxB=max(AE{n}(:,4))    % max benefit
    
    W=A.*E;                     % weighted adjacency matrix
    G=graph(W);                 % create graph
    
    % index and order nodes by characteristics
    IO=[nLabels' DC CC BC EC PR DS KC NM KG];
    ICDC=sortrows(IO(NC,[1 2]),2,'ascend');
    IDDC=sortrows(IO(ND,[1 2]),2,'ascend');
    UCDC=unique(ICDC(:,2));
    UDDC=unique(IDDC(:,2));
    ICCC=sortrows(IO(NC,[1 3]),2,'ascend');
    IDCC=sortrows(IO(ND,[1 3]),2,'ascend');
    ICBC=sortrows(IO(NC,[1 4]),2,'ascend');
    IDBC=sortrows(IO(ND,[1 4]),2,'ascend');
    ICEC=sortrows(IO(NC,[1 5]),2,'ascend');
    IDEC=sortrows(IO(ND,[1 5]),2,'ascend');
    ICPR=sortrows(IO(NC,[1 6]),2,'ascend');
    IDPR=sortrows(IO(ND,[1 6]),2,'ascend');
    ICDS=sortrows(IO(NC,[1 7]),2,'ascend');
    IDDS=sortrows(IO(ND,[1 7]),2,'ascend');
    ICKC=sortrows(IO(NC,[1 8]),2,'ascend');
    IDKC=sortrows(IO(ND,[1 8]),2,'ascend');
    UCKC=unique(ICKC(:,2));
    UDKC=unique(IDKC(:,2));
    ICNM=sortrows(IO(NC,[1 9]),2,'ascend');
    IDNM=sortrows(IO(ND,[1 9]),2,'ascend');
    UCNM=unique(ICNM(:,2));
    UDNM=unique(IDNM(:,2));
    ICKG=sortrows(IO(NC,[1 10]),2,'ascend');
    IDKG=sortrows(IO(ND,[1 10]),2,'ascend');
    UCKG=unique(ICKG(:,2));
    UDKG=unique(IDKG(:,2));
    
    tic
    % HILL CLIMBING ALGORITHM MAIN PROGRAM
    
    for r=1:rrr
        % initialize close and distant populations
        P=zeros(1,4);
        %     P(:,1)=randsample(NC,N);
        P(1)=randsample(ND,1);
        test=NC(find(E(P(1),NC)+DS(NC)'<=T));
        while isempty(test)==1
            P(1)=randsample(ND,1);
            test=NC(find(E(P(1),NC)+DS(NC)'<=T));
        end
        P(2)=test(randsample(length(test),1));
        
        % calculate costs and benefits
        P(3)=E(P(1),P(2));
        Ap=A;
        Ap(P(1),P(2))=1;
        Ap(P(2),P(1))=1;
        DSp=distances(graph(Ap.*E),nLabels,F);
        P(4)=length(find(DSp<=T))-length(find(DS<=T));
        
        Opt(1,:)=zeros(1,4);
        Opt(2,:)=P(1:4);
        
        traj=zeros(1,6);
        
        v=1;
        
        k=2;
        while v<5
            
            CB=0;
            
            if sum(Opt(k,:)==Opt(k-1,:))<=4
                v=v+1;
            else
                v=1;
            end
            
            
            % shuffle degree
            pos=find(IDDC(:,1)==P(1));
            pos2=find(UDDC==IDDC(pos,2));
            f=find(IDDC(:,2)==UDDC(pos2));
            ns=min(v,length(f));
            HD=[IDDC(f(randsample(length(f),ns))) ones(ns,1)];
            % decrease degree
            if IDDC(pos,2)>min(UDDC)
                f=find(find(IDDC(:,2)==UDDC(pos2-1)));
                ns=min(v,length(f));
                HD=[HD;IDDC(f(randsample(length(f),ns))) 2*ones(ns,1)];
            end
            % increase degree
            if IDDC(pos,2)<max(UDDC)
                f=find(find(IDDC(:,2)==UDDC(pos2+1)));
                ns=min(v,length(f));
                HD=[HD;IDDC(f(randsample(length(f),ns))) 3*ones(ns,1)];
            end
            % decrease distance
            pos=find(IDDS(:,1)==P(1));
            if pos>1
                ns=min(v,length(IDDS(1:(pos-1))));
                HD=[HD;IDDS((pos-ns):(pos-1))' 4*ones(ns,1)];
            end
            % increase distance
            if pos<mm
                ns=min(v,length(IDDS((pos+1):mm)));
                HD=[HD;IDDS((pos+1):(pos+ns))' 5*ones(ns,1)];
            end
            % decrease closeness
            pos=find(IDCC(:,1)==P(1));
            if pos>1
                ns=min(v,length(IDCC(1:(pos-1))));
                HD=[HD;IDCC((pos-ns):(pos-1))' 6*ones(ns,1)];
            end
            % increase closeness
            if pos<mm
                ns=min(v,length(IDCC((pos+1):mm)));
                HD=[HD;IDCC((pos+1):(pos+ns))' 7*ones(ns,1)];
            end
            % decrease betweenness
            pos=find(IDBC(:,1)==P(1));
            if pos>1
                ns=min(v,length(IDBC(1:(pos-1))));
                HD=[HD;IDBC((pos-ns):(pos-1))' 8*ones(ns,1)];
            end
            % increase betweenness
            if pos<mm
                ns=min(v,length(IDBC((pos+1):mm)));
                HD=[HD;IDBC((pos+1):(pos+ns))' 9*ones(ns,1)];
            end
            % decrease eigenvector
            pos=find(IDEC(:,1)==P(1));
            if pos>1
                ns=min(v,length(IDEC(1:(pos-1))));
                HD=[HD;IDEC((pos-ns):(pos-1))' 10*ones(ns,1)];
            end
            % increase eigenvector
            if pos<mm
                ns=min(v,length(IDEC((pos+1):mm)));
                HD=[HD;IDEC((pos+1):(pos+ns))' 11*ones(ns,1)];
            end
            % decrease pagerank
            pos=find(IDPR(:,1)==P(1));
            if pos>1
                ns=min(v,length(IDPR(1:(pos-1))));
                HD=[HD;IDPR((pos-ns):(pos-1))' 12*ones(ns,1)];
            end
            % increase pagerank
            if pos<mm
                ns=min(v,length(IDPR((pos+1):mm)));
                HD=[HD;IDPR((pos+1):(pos+ns))' 13*ones(ns,1)];
            end
            % decrease cluster coeff
            pos=find(IDKC(:,1)==P(1));
            if pos>1
                ns=min(v,length(IDKC(1:(pos-1))));
                HD=[HD;IDKC((pos-ns):(pos-1))' 14*ones(ns,1)];
            end
            % increase cluster coeff
            if pos<mm
                ns=min(v,length(IDKC((pos+1):mm)));
                HD=[HD;IDKC((pos+1):(pos+ns))' 15*ones(ns,1)];
            end
            % shuffle mod
            pos=find(IDNM(:,1)==P(1));
            pos2=find(UDNM==IDNM(pos,2));
            f=find(IDNM(:,2)==UDNM(pos2));
            ns=min(v,length(f));
            HD=[HD;IDNM(f(randsample(length(f),ns))) 16*ones(ns,1)];
            % decrease mod
            if IDNM(pos,2)>min(UDNM)
                f=find(IDNM(:,2)==UDNM(pos2-1));
                ns=min(v,length(f));
                HD=[HD;IDNM(f(randsample(length(f),ns))) 17*ones(ns,1)];
            end
            % increase mod
            if IDNM(pos,2)<max(UDNM)
                f=find(IDNM(:,2)==UDNM(pos2+1));
                ns=min(v,length(f));
                HD=[HD;IDNM(f(randsample(length(f),ns))) 18*ones(ns,1)];
            end
            % shuffle gap
            pos=find(IDKG(:,1)==P(1));
            pos2=find(UDKG==IDKG(pos,2));
            f=find(IDKG(:,2)==UDKG(pos2));
            ns=min(v,length(f));
            HD=[HD;IDKG(f(randsample(length(f),ns))) 19*ones(ns,1)];
            % decrease gap
            if IDKG(pos,2)>min(UDKG)
                f=find(IDKG(:,2)==UDKG(pos2-1));
                ns=min(v,length(f));
                HD=[HD;IDKG(f(randsample(length(f),ns))) 20*ones(ns,1)];
            end
            % increase gap
            if IDKG(pos,2)<max(UDKG)
                f=find(IDKG(:,2)==UDKG(pos2+1));
                ns=min(v,length(f));
                HD=[HD;IDKG(f(randsample(length(f),ns))) 21*ones(ns,1)];
            end
            % neighbors
            GN=neighbors(G,P(1));
            GN=setdiff(GN,NC);
            if isempty(GN)==0
                GN=[GN 22*ones(length(GN),1)];
                HD=[HD; GN];
            end
            
            
            % shuffle degree
            pos=find(ICDC(:,1)==P(2));
            pos2=find(UCDC==ICDC(pos,2));
            f=find(ICDC(:,2)==UCDC(pos2));
            ns=min(v,length(f));
            HC=[ICDC(f(randsample(length(f),ns))) ones(ns,1)];
            % decrease degree
            if ICDC(pos,2)>min(UCDC)
                f=find(find(ICDC(:,2)==UCDC(pos2-1)));
                ns=min(v,length(f));
                HC=[HC;ICDC(f(randsample(length(f),ns))) 2*ones(ns,1)];
            end
            % increase degree
            if ICDC(pos,2)<max(UCDC)
                f=find(find(ICDC(:,2)==UCDC(pos2+1)));
                ns=min(v,length(f));
                HC=[HC;ICDC(f(randsample(length(f),ns))) 3*ones(ns,1)];
            end
            % decrease distance
            pos=find(ICDS(:,1)==P(2));
            if pos>1
                ns=min(v,length(ICDS(1:(pos-1))));
                HC=[HC;ICDS((pos-ns):(pos-1))' 4*ones(ns,1)];
            end
            % increase distance
            if pos<mm
                ns=min(v,length(ICDS((pos+1):mm)));
                HC=[HC;ICDS((pos+1):(pos+ns))' 5*ones(ns,1)];
            end
            % decrease closeness
            pos=find(ICCC(:,1)==P(2));
            if pos>1
                ns=min(v,length(ICCC(1:(pos-1))));
                HC=[HC;ICCC((pos-ns):(pos-1))' 6*ones(ns,1)];
            end
            % increase closeness
            if pos<mm
                ns=min(v,length(ICCC((pos+1):mm)));
                HC=[HC;ICCC((pos+1):(pos+ns))' 7*ones(ns,1)];
            end
            % decrease betweenness
            pos=find(ICBC(:,1)==P(2));
            if pos>1
                ns=min(v,length(ICBC(1:(pos-1))));
                HC=[HC;ICBC((pos-ns):(pos-1))' 8*ones(ns,1)];
            end
            % increase betweenness
            if pos<mm
                ns=min(v,length(ICBC((pos+1):mm)));
                HC=[HC;ICBC((pos+1):(pos+ns))' 9*ones(ns,1)];
            end
            % decrease eigenvector
            pos=find(ICEC(:,1)==P(2));
            if pos>1
                ns=min(v,length(ICEC(1:(pos-1))));
                HC=[HC;ICEC((pos-ns):(pos-1))' 10*ones(ns,1)];
            end
            % increase eigenvector
            if pos<mm
                ns=min(v,length(ICEC((pos+1):mm)));
                HC=[HC;ICEC((pos+1):(pos+ns))' 11*ones(ns,1)];
            end
            % decrease pagerank
            pos=find(ICPR(:,1)==P(2));
            if pos>1
                ns=min(v,length(ICPR(1:(pos-1))));
                HC=[HC;ICPR((pos-ns):(pos-1))' 12*ones(ns,1)];
            end
            % increase pagerank
            if pos<mm
                ns=min(v,length(ICPR((pos+1):mm)));
                HC=[HC;ICPR((pos+1):(pos+ns))' 13*ones(ns,1)];
            end
            % decrease cluster coeff
            pos=find(ICKC(:,1)==P(2));
            if pos>1
                ns=min(v,length(ICKC(1:(pos-1))));
                HC=[HC;ICKC((pos-ns):(pos-1))' 14*ones(ns,1)];
            end
            % increase cluster coeff
            if pos<mm
                ns=min(v,length(ICKC((pos+1):mm)));
                HC=[HC;ICKC((pos+1):(pos+ns))' 15*ones(ns,1)];
            end
            % shuffle mod
            pos=find(ICNM(:,1)==P(2));
            pos2=find(UCNM==ICNM(pos,2));
            f=find(ICNM(:,2)==UCNM(pos2));
            ns=min(v,length(f));
            HC=[HC;ICNM(f(randsample(length(f),ns))) 16*ones(ns,1)];
            % decrease mod
            if ICNM(pos,2)>min(UCNM)
                f=find(ICNM(:,2)==UCNM(pos2-1));
                ns=min(v,length(f));
                HC=[HC;ICNM(f(randsample(length(f),ns))) 17*ones(ns,1)];
            end
            % increase mod
            if ICNM(pos,2)<max(UCNM)
                f=find(ICNM(:,2)==UCNM(pos2+1));
                ns=min(v,length(f));
                HC=[HC;ICNM(f(randsample(length(f),ns))) 18*ones(ns,1)];
            end
            % shuffle gap
            pos=find(ICKG(:,1)==P(2));
            pos2=find(UCKG==ICKG(pos,2));
            f=find(ICKG(:,2)==UCKG(pos2));
            ns=min(v,length(f));
            HC=[HC;ICKG(f(randsample(length(f),ns))) 19*ones(ns,1)];
            % decrease gap
            if ICKG(pos,2)>min(UCKG)
                f=find(ICKG(:,2)==UCKG(pos2-1));
                ns=min(v,length(f));
                HC=[HC;ICKG(f(randsample(length(f),ns))) 20*ones(ns,1)];
            end
            % increase gap
            if ICKG(pos,2)<max(UCKG)
                f=find(ICKG(:,2)==UCKG(pos2+1));
                ns=min(v,length(f));
                HC=[HC;ICKG(f(randsample(length(f),ns))) 21*ones(ns,1)];
            end
            % neighbors
            GN=neighbors(G,P(2));
            GN=setdiff(GN,NC);
            if isempty(GN)==0
                GN=[GN 22*ones(length(GN),1)];
                HC=[HC; GN];
            end
            
            % calculate costs and benefits
            HD=[P(1) 0; HD];
            HC=[P(2) 0; HC];
            HD(~any(HD(:,1),2),:)=[];
            HC(~any(HC(:,1),2),:)=[];
            %             HD=unique(HD,'rows');
            %             HC=unique(HC,'rows');
            kk=1;
            kkk=1;
            for j=HD(:,1)'
                SS=HC(find(E(j,HC(:,1))+DS(HC(:,1))'<=T),1)';
                kkkk=1;
                for jj=SS
                    CB(kk,1)=j;
                    CB(kk,2)=jj;
                    CB(kk,3)=E(j,jj);
                    Ap=A;
                    Ap(j,jj)=1;
                    Ap(jj,j)=1;
                    DSp=distances(graph(Ap.*E),nLabels,F);
                    CB(kk,4)=length(find(DSp<=T))-length(find(DS<=T));
                    CB(kk,5)=HD(kkk,2);
                    CB(kk,6)=HC(kkkk,2);
                    kk=kk+1;
                    kkkk=kkkk+1;
                end
                kkk=kkk+1;
            end
            
            P=CB(find(CB(:,4)==max(CB(:,4))),:);
            if length(P(:,1))>1
                P=P(find(P(:,3)==min(P(:,3))),:);
                P=P(length(P(:,1)),:);
            end
            
            traj(k-1,:)=P;
            
            k=k+1;
            
            Opt(k,:)=P(1:4);
            
        end
        
        t=toc;
        
        CB=P;
        o=CB(find(CB(:,4)==max(CB(:,4))),:);
        if length(o)>1
            o=o(find(o(:,3)==min(o(:,3))),:);
            o=o(1,:);
        end
        
        O(rr,:)=[t o(1) o(2) o(3) o(4) DC(o(1)) DC(o(2)) CC(o(1)) CC(o(2)) BC(o(1)) BC(o(2)) EC(o(1)) EC(o(2)) PR(o(1)) PR(o(2)) DS(o(1)) DS(o(2)) KC(o(1)) KC(o(2)) NM(o(1)) NM(o(2)) KG(o(1)) KG(o(2))];
        if (rr>1) && (mod(rr-1,20)~=0)
            O(rr,1)=t-O(rr-1,1);
        end
        
        [rr O(rr,1) o(3) o(4)]
        
        Traj{rr}=traj;
        
        rr=rr+1;
        
    end
    
    %     AE{n}=[CB DC(CB(:,1)) DC(CB(:,2)) CC(CB(:,1)) CC(CB(:,2)) BC(CB(:,1)) BC(CB(:,2)) EC(CB(:,1)) EC(CB(:,2)) PR(CB(:,1)) PR(CB(:,2)) DS(CB(:,1)) DS(CB(:,2))];
    
    
end


MO=zeros(ln,6);
for i=1:ln
    MO(i,:)=[mean(O((rrr*(i-1)+1):(rrr*i),1)) var(O((rrr*(i-1)+1):(rrr*i),1)) mean(O((rrr*(i-1)+1):(rrr*i),4)) var(O((rrr*(i-1)+1):(rrr*i),4)) mean(O((rrr*(i-1)+1):(rrr*i),5)) var(O((rrr*(i-1)+1):(rrr*i),5))];
end

save(['er_',num2str(N),'_HCVN.mat'],'O','MO','Traj')




